@@ -9,8 +9,12 @@
 #include "py.h"
 #include <cuda.h>
 
+struct DriverApi;
+
 Status cuda_helper_init(PyObject* m);
 
-const char* get_cuda_error(CUresult res);
+const char* get_cuda_error(const DriverApi*, CUresult res);
 
-void try_init_cuda();
+void try_cuInit(const DriverApi*);
+
+Status check_driver_version(const DriverApi*, int minimum_version);
